from devicemanager.connectionmanager import *
from devicemanager.screenmanager import *
from devicemanager.gpsmanager import *
from devicemanager.systemmanager import *
from gtkeasyapp.log import *
from gtkeasyapp.gis import *
from foxytag import *
import threading
import copy

##FoxyTagForSpeedcamAlerter contains all data needed by speedcam alerter
class FoxyTagForSpeedcamAlerter():
	##Constructor
	def __init__(self, foxyTag):
		#init members
		self.__foxyTag = None
		self.__utmPoint = None
		self.__wgs84Point = WGS84Point()
		self.__distanceFromCar = 100000 #distance between car and speedcam
		
		#set foxy tag
		self.__setFoxyTag(foxyTag)
	
	##Return foxy tag
	def getFoxyTag(self):
		return self.__foxyTag
	
	##Set foxy tag and compute utm position
	def __setFoxyTag(self, foxyTag):
		#set member
		self.__foxyTag = foxyTag
		
		#save foxy tag wgs84 position
		self.__wgs84Point.latitude = foxyTag.lat
		self.__wgs84Point.longitude = foxyTag.long
					
		#compute foxytag utm position
		self.__utmPoint = GISConvert.wgs84ToUtm(self.__wgs84Point)
	
	##Compute and return distance between car and speedcam
	def getDistanceFromCar(self, utmCarPosition=None):
		#compute distance
		if utmCarPosition!=None:
			self.__distanceFromCar = GISCompute.distanceUtm(utmCarPosition, self.__utmPoint)
		
		#return computed distance
		return self.__distanceFromCar
	
	##Return foxy tag UTM position
	def getUtmPoint(self):
		return self.__utmPoint
	
	##Return foxy tag WGS84 position
	def getWgs84Ppoint(self):
		return self.__wgs84Point
	
	##Return true if speedcam is in front of the car
	def isInCarDirection(self):
		return not self.__foxyTag.opposite

class SpeedcamAlerter():
	##constants
	GPS_ACQUIREMENT = 1.0 #in seconds
	FOXY_TIMEOUT_MAX = 270.0 #in seconds (4min 30secs)
	FOXY_RADIUS = 7250 #in meters (best compromise according to foxy documentation)
	PRECISION_POSITION = 35 #meters
	PRECISION_HEADING = 7 #degrees
	PRECISION_SPEED = 7 #km/h

	##Constructor
	def __init__(self, pref, setSpeedCallback, setDistanceCallback, setMessageCallback, setPositionCallback, addFoxyTagCallback, removeAllFoxyTagsCallback, setSpeedcamIsCloseCallback):		
		#init members
		self.__resetMembers()
		self.__pref = pref
		
		#init managers
		self.__screen = ScreenManager()
		self.__connections = ConnectionManager()
		self.__gps = GpsManager()
		self.__foxy = Foxy()
		
		#init callbacks
		self.__setSpeedCallback = setSpeedCallback
		self.__setDistanceCallback = setDistanceCallback
		self.__setMessageCallback = setMessageCallback
		self.__setPositionCallback = setPositionCallback
		self.__addFoxyTagCallback = addFoxyTagCallback
		self.__removeAllFoxyTagsCallback = removeAllFoxyTagsCallback
		self.__setSpeedcamIsCloseCallback = setSpeedcamIsCloseCallback
		
		#screen always on
		self.__connection = None
		connection = self.__pref.getItem('connections', 'connection')
		if connection!=None and len(connection.getValue())>0:
			connections = self.__connections.getConnections(ConnectionItem.CNX_TYPE_DATA, connection.getValue())
			if len(connections)>0:
				self.__connection = connections[0]
				Log.log('%s data connections found' % str(self.__connection))
	
	##Reset all members
	def __resetMembers(self):
		self.__gpsTimer = None
		self.__foxyTimer = None
		self.__displayOnTimer = None
		self.__stopped = True
		self.__foxyInit = False
		self.__foxyTags = []
		self.__currentPositionWgs84 = None
		self.__currentPositionUtm = None
		self.__currentSpeed = 0 #in km/h
		self.__currentAngle = 0 #0=360=north; 180=south
		self.__utmPositionFromLastFoxytagsUpdate = None
		self.__previousDistance = 100000
		self.__createConnection = False
		self.__gpsAcquirement = 1.0 #1 second
		self.__actionPosition = None
		
	##Return True if speedcamalerter is stopped
	def isStopped(self):
		return self.__stopped

	##Start speedcam alerter
	def start(self):
		#init members
		self.__resetMembers()
		self.__currentPositionWgs84 = WGS84Point()
		self.__stopped = False
		
		#screen always on
		eco = self.__pref.getItem('display', 'eco')
		if eco!=None and not eco.getValue():
			#turn screen on every 55 seconds
			self.__displayOnTimer = threading.Timer(55.0, self.__turnScreenOn, [True])
			self.__displayOnTimer.start()
	
		#start GPS
		self.__gpsTimer = threading.Timer(2.0, self.__startGps)
		self.__gpsTimer.start()
		
		#log
		Log.log('SpeedcamAlerter started')
	
	##Start GPS
	def __startGps(self):
		if not self.isStopped():
			#start gps
			try:
				#start gps
				self.__gps.start(self.__onGpsStart, self.__onGpsNeedRestart, self.__onGpsFix)
				
			except GpsManagerException, e:
				#unable to start GPS
				Log.log('GpsManagerException : %s' % str(e))
				#stop all
				self.stop()
		
		return False
	
	##Gps is started
	def __onGpsStart(self):
		#log message
		Log.log('GPS started')
	
	##Gps needs to be restarted
	def __onGpsNeedRestart(self):
		#log message
		Log.log('GPS needs to be restarted')
		
		#stop speedcamalerter
		self.stop()
		time.sleep(2.0)
		self.start()

	##Gps has fix
	def __onGpsFix(self):
		#log message
		Log.log('GPS is fixed')
		
		#launch data acquirement every seconds
		if not self.__stopped:
			self.__gpsTimer = threading.Timer(self.__gpsAcquirement, self.__getGpsData)
			self.__gpsTimer.start()
		
		#get foxy data
		if not self.__stopped:
			self.__foxyTimer = threading.Timer(1.0, self.__connectToFoxy)
			self.__foxyTimer.start()
			
	##Stop speedcam alerter
	def stop(self):
		#stop gps device
		self.__gps.stop()
	
		#stop connection
		if self.__createConnection:
			self.__connections.disconnect()
		
		#stop threads
		if self.__gpsTimer!=None:
			self.__gpsTimer.cancel()
			self.__gpsTimer = None
		if self.__foxyTimer!=None:
			self.__foxyTimer.cancel()
			self.__foxyTimer = None
		if self.__displayOnTimer!=None:
			self.__displayOnTimer.cancel()
			self.__displayOnTimer = None
			
		#reset members
		self.__resetMembers()
		self.__stopped = True
		
		#log
		Log.log('SpeedcamAlerter stopped')
	
	##Reset position saved for foxytag action (add/remove speedcam)
	def resetActionPosition(self):
		self.__actionPosition = None
	
	##Save position saved for foxytag action (add/remove speedcam)
	def saveActionPosition(self):
		#get current gps position
		data = self.__gps.getData()
		
		if data!=None:
			self.__actionPosition = WGS84Point(data.latitude, data.longitude, data.eph, data.eps, data.epd)
			Log.log('save action position : %s' % self.__actionPosition)
		elif self.__currentPositionWgs84!=None:
			#unable to get current gps position, use position in use in speedcamalerter
			self.__actionPosition = WGS84Point()
			self.__actionPosition.copy(self.__currentPositionWgs84)
			Log.log('save action position : %s' % self.__actionPosition)
		else:
			#nothing available
			pass
	
	##Cancel speedcam
	def removeSpeedcam(self):
		#post tag
		threading.Thread(None, self.__postFoxyTag, None, (FoxyTag.FOXYTAG_CANCEL)).start()
	
	##Add fixed speedcam
	def addFixedSpeedcam(self):
		#post tag
		threading.Thread(None, self.__postFoxyTag, None, (FoxyTag.FOXYTAG_FIXED, False)).start()
	
	##Add opposite fixed speedcam
	def addOppositeFixedSpeedcam(self):
		#post tag
		threading.Thread(None, self.__postFoxyTag, None, (FoxyTag.FOXYTAG_FIXED, True)).start()
	
	##Cancel mobile speedcam
	def addMobileSpeedcam(self):
		#post tag
		threading.Thread(None, self.__postFoxyTag, None, (FoxyTag.FOXYTAG_MOBILE, False)).start()
	
	##Cancel opposite mobile speedcam
	def addOppositeMobileSpeedcam(self):
		#post tag
		threading.Thread(None, self.__postFoxyTag, None, (FoxyTag.FOXYTAG_MOBILE, True)).start()
	
	##Post foxytag (add or confirm)
	##@param foxyTagKind : kind of foxy tag to post
	##@param opposite : False if speedcam isn't in car way, True otherwise
	def __postFoxyTag(self, foxyTagKind, opposite=False):
		if not self.__stopped:
			if self.__actionPosition!=None:
				#create new foxy tag
				tag = FoxyTag()
				tag.lat = self.__actionPosition.latitude
				tag.long = self.__actionPosition.longitude
				tag.kind = foxyTagKind
				tag.heading = self.__currentAngle
				if opposite:
					tag.heading += 1180
		
				#check if precision is good
				if self.__actionPosition.eph<=self.PRECISION_POSITION and self.__actionPosition.epd<=self.PRECISION_HEADING and self.__actionPosition.eps<=self.PRECISION_SPEED:
					#ok, post tag
					res, usize, dsize = self.__foxy.tagPost(tag, self.__currentSpeed)
				
					#update upload and download
					upload = self.__pref.getItem('connections', 'upload')
					upload.setValue( upload.getValue() + usize )
					download = self.__pref.getItem('connections', 'download')
					download.setValue( download.getValue() + dsize )
					self.__pref.save()
				else:
					Log.log('Precision is not sufficient to post foxytag (eph=%f, epd=%d, eps=%s)' % (self.__actionPosition.eph, self.__actionPosition.epd, self.__actionPosition.eps))
				
				#reset action position
				self.resetActionPosition()
			else:
				Log.log('No action position, unable to post foxytag')
		else:
			Log.log('SpeedcamAlerter stopped, unable to post foxytag')
		
	##Connect to foxy server
	def __connectToFoxy(self, status=-1, error=-1, iap_id=-1, bearer=-1):
		if not self.__stopped:
			#log
			Log.log('Connecting to internet... (status=%s error=%s iap=%s bearer=%s)' % (status, error, iap_id, bearer))
			
			#cancel current foxy timer if necessary
			if self.__foxyTimer!=None:
				self.__foxyTimer.cancel()
			else:
				Log.log('Max timeout reached, update foxytags now')
			
			#check if connection is available
			if self.__connections.isConnected():
				#connected
				
				#init connection to foxy
				foxyRes = True
				if not self.__foxyInit:
					#connect to foxytag
					username = self.__pref.getItem('connections', 'username')
					if username!=None:
						username = username.getValue()
					else:
						username = ''
					password = self.__pref.getItem('connections', 'password')
					if password!=None:
						password = password.getValue()
					else:
						password = ''
					foxyRes, usize, dsize, messages = self.__foxy.connect(username, password)

					if foxyRes:
						#update foxy init flag
						self.__foxyInit = True
					
						#update upload and download counters
						upload = self.__pref.getItem('connections', 'upload')
						upload.setValue( upload.getValue() + usize )
						download = self.__pref.getItem('connections', 'download')
						download.setValue( download.getValue() + dsize )
						self.__pref.save()
					else:
						#problem connecting to foxy server, retry in 30 secs
						timeout = 30
						Log.log('Unable to connect to foxytag server, restart in %d seconds' % (timeout))
					
					#display message
					if len(messages)>0:
						#display only first message
						self.__setMessageCallback(messages[0])
						#turn screen on
						self.__turnScreenOn()

				#get foxy tags
				if foxyRes:
					Log.log('Connected, getting foxy tags...')
					timeout,disconnect = self.__getFoxyTags()
				
					#update last foxytags update position
					self.__utmPositionFromLastFoxytagsUpdate = copy.copy(self.__currentPositionUtm)
				
					#disconnect if necessary
					if disconnect:
						self.__connections.disconnect()
				
				#launch again internet connection timer
				if not self.__stopped:
					#Log.log('Foxy timeout = %d' % timeout)
					self.__foxyTimer = threading.Timer(timeout, self.__connectToFoxy)
					self.__foxyTimer.start()
			else:
				if self.__connection!=None:
					#not connected, launch connection
					Log.log('Not connected to internet, trying to connect...')
					self.__connections.connectTo(self.__connection, self.__connectToFoxy)
				
					#speedcamalerter create connection
					self.__createConnection = True
				else:
					#no connection specified, nothing to do
					Log.log('No connection specified, unable to connect to internet')
		
	##Get foxy tags
	def __getFoxyTags(self):
		timeout = self.FOXY_TIMEOUT_MAX #request foxytag every 5 mins by default
		disconnect = False
		
		if self.__currentPositionWgs84!=None and self.__currentPositionWgs84.latitude!=0.0 and self.__currentPositionWgs84.longitude!=0.0:
			#current position seems to be correct
			radius = self.FOXY_RADIUS
			Log.log('Requesting foxytags for position %s with radius of %s' % (str(self.__currentPositionWgs84), str(radius)))
			foxyRes, usize, dsize, messages = self.__foxy.tagRequest(self.__currentPositionWgs84.latitude, self.__currentPositionWgs84.longitude, radius)
			if foxyRes:
				#ok, tag requested successfully
				
				#get found foxytags
				foxyTags = self.__foxy.getTags()
				
				#update radar tags
				Log.log('Add %d new foxy tags' % len(foxyTags))
				if self.__removeAllFoxyTagsCallback!=None:
					self.__removeAllFoxyTagsCallback()
					if self.__addFoxyTagCallback!=None:
						for foxytag in foxyTags:
							self.__addFoxyTagCallback(foxytag)
				#Log.log('End adding foxytags')
				
				#save foxy tag in speedcam alerter useful format
				for foxyTag in foxyTags:
					#create new object
					self.__foxyTags.append(FoxyTagForSpeedcamAlerter(foxyTag))
				
				#tags retrieved successfully, stay connected and return max timeout
				disconnect = False
				timeout = self.FOXY_TIMEOUT_MAX
				
				#update upload and download
				upload = self.__pref.getItem('connections', 'upload')
				upload.setValue( upload.getValue() + usize )
				download = self.__pref.getItem('connections', 'download')
				download.setValue( download.getValue() + dsize )
				self.__pref.save()
				
			else:
				#unable to get foxytags, nothing to do
				Log.log('Unable to get foxy tags, disconnect and retry')
				disconnect = True
				timeout = 15.0
				
			#display message
			if len(messages)>0:
				#display only first message
				self.__setMessageCallback(messages[0])
				#turn screen on
				self.__turnScreenOn()
				
		else:
			#current position seems to be invalid, launch again in 5 seconds (time to get gps fix?)
			Log.log('Current position seems to be invalid, don\'t disconnect and retry in 5 seconds')
			disconnect = False
			timeout = 5.0
		
		return timeout, disconnect
	
	##Check if speedcam is in car direction
	##@param speedcamAngle : speedcam angle
	def __isSpeedcamInCarDirection(self, speedcamAngle):
		#init
		result = True
		debug = 1
		minAngleThreshold = 0
		maxAngleThreshold = 0
	
		if speedcamAngle!=-1:
			flag = False
		
			#compute min angle threshold
			minAngleThreshold = self.__currentAngle - 45
			if minAngleThreshold<0:
				minAngleThreshold += 360
				flag = True
			
			#compute min angle threshold
			maxAngleThreshold = self.__currentAngle + 45
			if maxAngleThreshold>360:
				maxAngleThreshold -= 360
				flag = True
				
			#detect if speedcam is in front of the car or in the opposite direction
			if flag:
				if speedcamAngle>=minAngleThreshold or speedcamAngle<=maxAngleThreshold:
					#speedcam seems to be in car direction
					result = True
					debug = 1
				else:
					#speedcam seems not to be in car direction
					result = False
					debug = 2
			else:
				if speedcamAngle>=minAngleThreshold and speedcamAngle<=maxAngleThreshold:
					#speedcam seems to be in car direction
					result = True
					debug = 3
				else:
					#speedcam seems not to be in car direction
					result = False
					debug = 4
		else:
			#no speedcam direction specified
			result = True
			debug = 5
	
		#Log.log('OPP : angle=%.2f min=%.2f heading=%.2f max=%.2f opp=%s dbg=%d' % (self.__currentAngle, minAngleThreshold, speedcamAngle, maxAngleThreshold, result, debug))
			
		return result
		
	##Turn screen on
	def __turnScreenOn(self, timer=False):
		if self.__screen!=None:
			self.__screen.turnScreenOnDuring60Seconds()
			
		#launch again if timer
		if timer:
			self.__displayOnTimer = threading.Timer(55.0, self.__turnScreenOn, [True])
			self.__displayOnTimer.start()
			
	##Set gps acquirement time
	#def __setGpsAcquirement(self, speedcamIsClosed=False):
	#	if speedcamIsClosed:
	#		self.__gpsAcquirement = 0.25
	#	elif self.__ecoMode:		
			
	##Get gps data
	def __getGpsData(self):
		#get data
		data = self.__gps.getData()
	
		if not self.__stopped and self.__currentPositionWgs84!=None:
			if data!=None and data.speed>0.556:
				#data available
				if data.mode!=GpsManager.MODE_NO_FIX:
					#fix ok, save current position
					self.__currentPositionWgs84.latitude = data.latitude
					self.__currentPositionWgs84.longitude = data.longitude
					self.__currentPositionWgs84.eph = data.eph
					self.__currentPositionWgs84.epd = data.epd
					self.__currentPositionWgs84.eps = data.eps
					print self.__currentPositionWgs84
					self.__currentPositionUtm = GISConvert.wgs84ToUtm(self.__currentPositionWgs84)
					
					#need to update foxytags ?
					if self.__currentPositionUtm!=None and self.__utmPositionFromLastFoxytagsUpdate!=None:
						if GISCompute.distanceUtm(self.__utmPositionFromLastFoxytagsUpdate, self.__currentPositionUtm)>=self.FOXY_RADIUS:
							Log.log('Max radius reached, update foxytags now')
							self.__connectToFoxy()
					
					#set current speed (in km/h)
					if self.__setSpeedCallback!=None and data.speed!=GpsManager.NaN:
						self.__currentSpeed = float(data.speed) * 3.6
						self.__setSpeedCallback(self.__currentSpeed)
					
					#set current position to radar
					if self.__setPositionCallback!=None:
						#gps heading
						self.__currentAngle = data.track
						self.__setPositionCallback(self.__currentPositionUtm, self.__currentSpeed, self.__currentAngle)
						
					#set current distance and direction from all speedcams
					#compute speedcam closest distance and speedcam direction according to car direction
					if self.__setDistanceCallback!=None and len(self.__foxyTags)>0:
						#init
						closestSpeedcam = None
						
						#walk through all stored foxy tags
						for foxyTag in self.__foxyTags:
							if closestSpeedcam==None:
								#set closest speedcam
								closestSpeedcam = foxyTag
							else:
								#compare closest speedcam to current one
								if closestSpeedcam.getDistanceFromCar(self.__currentPositionUtm)>foxyTag.getDistanceFromCar(self.__currentPositionUtm):
									#the current one is the closest
									closestSpeedcam = foxyTag
							
							#compute speedcam direction if necessary
							foxyTag.getFoxyTag().opposite = not self.__isSpeedcamInCarDirection(foxyTag.getFoxyTag().heading)
						
						#here we should know the closest speedcam
						
						#update closest distance
						self.__setDistanceCallback(closestSpeedcam)
						
						#get distance alert
						if closestSpeedcam.getFoxyTag().kind==FoxyTag.FOXYTAG_FIXED:
							key = 'fixed'
						elif closestSpeedcam.getFoxyTag().kind==FoxyTag.FOXYTAG_MOBILE:
							key = 'mobile'
						else:
							key = 'ghost'
						alert = self.__pref.getItem('general', key)
						if alert==None:
							distAlert = 1000 #default alert is 1km
						else:
							distAlert = alert.getValue()
						
						#check if there is something to do
						if closestSpeedcam.getDistanceFromCar()<distAlert and closestSpeedcam.isInCarDirection():
							#a speedcam is closest to the car and in front of the car
							#change gps acquirement time
							self.__gpsAcquirement = 0.25
							
							#update radar
							self.__setSpeedcamIsCloseCallback(True, closestSpeedcam.getDistanceFromCar())
							
							#turn on screen
							if self.__previousDistance>closestSpeedcam.getDistanceFromCar():
								self.__turnScreenOn()
							self.__previousDistance = closestSpeedcam.getDistanceFromCar()
						else:
							#change gps acquirement time
							self.__gpsAcquirement = 1.0
						
							#nothing to do, just say to radar that there is no speedcam close to the car
							self.__setSpeedcamIsCloseCallback(False, closestSpeedcam.getDistanceFromCar())
				else:
					#no fix
					#Log.log('No GPS fix')
					
					#change gps acquirement time
					self.__gpsAcquirement = 1.0
					
					#reset position
					self.__currentPositionWgs84.latitude = 0.0
					self.__currentPositionWgs84.longitude = 0.0
					
					#reset current distance from nearest speedcam
					if self.__setDistanceCallback!=None:
						self.__setDistanceCallback(0)
					
					#reset speed
					if self.__setSpeedCallback!=None:
						self.__setSpeedCallback(0)
			else:
				#no data, strange?
				#Log.log('No GPS data or speed is too low (%d)' % data.speed)
				self.__currentPositionWgs84.latitude = 0.0
				self.__currentPositionWgs84.longitude = 0.0
				
				#change gps acquirement time
				self.__gpsAcquirement = 1.0
	
		#launch again
		if not self.__stopped:
			self.__gpsTimer = threading.Timer(self.__gpsAcquirement, self.__getGpsData)
			self.__gpsTimer.start()
